@@ -3,8 +3,12 @@
 from __future__ import annotations as _annotations
 
 import re
+from typing import Any, ClassVar, Literal
 
-from typing_extensions import Literal, Self
+from typing_extensions import Self
+from typing_inspection.introspection import Qualifier
+
+from pydantic._internal import _repr
 
 from ._migration import getattr_migration
 from .version import version_short
@@ -15,6 +19,7 @@ __all__ = (
     'PydanticImportError',
     'PydanticSchemaGenerationError',
     'PydanticInvalidForJsonSchema',
+    'PydanticForbiddenQualifier',
     'PydanticErrorCodes',
 )
 
@@ -71,6 +76,7 @@ PydanticErrorCodes = Literal[
     'unpack-typed-dict',
     'overlapping-unpack-typed-dict',
     'invalid-self-type',
+    'validate-by-alias-and-name-false',
 ]
 
 
@@ -159,4 +165,26 @@ class PydanticInvalidForJsonSchema(PydanticUserError):
         super().__init__(message, code='invalid-for-json-schema')
 
 
+class PydanticForbiddenQualifier(PydanticUserError):
+    """An error raised if a forbidden type qualifier is found in a type annotation."""
+
+    _qualifier_repr_map: ClassVar[dict[Qualifier, str]] = {
+        'required': 'typing.Required',
+        'not_required': 'typing.NotRequired',
+        'read_only': 'typing.ReadOnly',
+        'class_var': 'typing.ClassVar',
+        'init_var': 'dataclasses.InitVar',
+        'final': 'typing.Final',
+    }
+
+    def __init__(self, qualifier: Qualifier, annotation: Any) -> None:
+        super().__init__(
+            message=(
+                f'The annotation {_repr.display_as_type(annotation)!r} contains the {self._qualifier_repr_map[qualifier]!r} '
+                f'type qualifier, which is invalid in the context it is defined.'
+            ),
+            code=None,
+        )
+
+
 __getattr__ = getattr_migration(__name__)
